#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <GL/glew.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <SDL.h>

#include "metaballs.h"
#include "meshgen.h"
#include "arcball.h"
#include "shader.h"

#define WIDTH  600
#define HEIGHT 600
#define BPP    24

static metaball_t balls[NR_METABALLS];
static int polygonMode = 1;
static GLuint em_cube_texture;

static const char *em_bmps[] = {
	"images/em_fr.bmp", "images/em_bk.bmp", 
	"images/em_dn.bmp", "images/em_up.bmp", 
	"images/em_lt.bmp", "images/em_rt.bmp"
};

/* Thanks to google ;) */
static float skybox_verts[24][3] =
{
	//Right side of sky box.xpos
	{	 10.0f,	 10.0f,	-10.0f	},	//Top left.
	{	 10.0f,	 10.0f,	 10.0f	},	//Top right.
	{	 10.0f,	-10.0f,	 10.0f	},	//Bottom right.
	{	 10.0f,	-10.0f,	-10.0f	},	//Bottom left.

	//left side of sky box.xneg
	{	-10.0f,	 10.0f,	 10.0f	},	//Top left.
	{	-10.0f,	 10.0f,	-10.0f	},	//Top right.
	{	-10.0f,	-10.0f,	-10.0f	},	//Bottom right.
	{	-10.0f,	-10.0f,	 10.0f	},	//Bottom left.

	//Top side of sky box.ypose
	{	-10.0f,	 10.0f,	 10.0f	},	//Top left.
	{	 10.0f,	 10.0f,	 10.0f	},	//Top right.
	{	 10.0f,	 10.0f,	-10.0f	},	//Bottom right.
	{	-10.0f,	 10.0f,	-10.0f	},	//Bottom left.

	//Bottom side of sky box.yneg
	{	-10.0f,	-10.0f,	-10.0f	},	//Top left.
	{	 10.0f,	-10.0f,	-10.0f	},	//Top right.
	{	 10.0f,	-10.0f,	 10.0f	},	//Bottom right.
	{	-10.0f,	-10.0f,	 10.0f	},	//Bottom left.

	//Front side of sky box.zpos
	{	 10.0f,	 10.0f,	 10.0f	},	//Top left.
	{	-10.0f,	 10.0f,	 10.0f	},	//Top right.
	{	-10.0f,	-10.0f,	 10.0f	},	//Bottom right.
	{	 10.0f,	-10.0f,	 10.0f	},	//Bottom left.

	//Back side of sky box.zneg
	{	-10.0f,	 10.0f,	-10.0f	},	//Top left.
	{	 10.0f,	 10.0f,	-10.0f	},	//Top right.
	{	 10.0f,	-10.0f,	-10.0f	},	//Bottom right.
	{	-10.0f,	-10.0f,	-10.0f	},	//Bottom left.
};

static GLuint shaderHandle;
static GLuint uniformTexUnit;

void setup_shader()
{
	shaderHandle = createShaderProgram("glsl/em.vert", "glsl/em.frag");
	uniformTexUnit = glGetUniformLocationARB(shaderHandle, "cubeMap");
}

void load_cube_em()
{
	SDL_Surface *img;
	int i;
	
	glGenTextures(1, &em_cube_texture);
	glBindTexture(GL_TEXTURE_CUBE_MAP, em_cube_texture);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);

	for (i = 0; i < 6; i++) {
		img = SDL_LoadBMP(em_bmps[i]);
		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i,
				0, 3, img->w, img->h, 0, GL_BGR, GL_UNSIGNED_BYTE, 
				img->pixels);

		SDL_FreeSurface(img);
	}

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_CUBE_MAP, em_cube_texture);
	
	
	glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
	glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
	glTexGenf(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);

	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);
	glEnable(GL_TEXTURE_GEN_R);

	glEnable(GL_TEXTURE_CUBE_MAP);
}
/*
void load_texture_2d(GLuint *tex, const char *filename)
{
	SDL_Surface *img;
	
	glGenTextures(1, tex);
	glBindTexture(GL_TEXTURE_2D, *tex);
	
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
			GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
		
	img = SDL_LoadBMP(filename);
	glTexImage2D(GL_TEXTURE_2D, 0, 3, img->w, img->h, 0, GL_BGR, 
			GL_UNSIGNED_BYTE, img->pixels);

	SDL_FreeSurface(img);
}
*/
void render_skybox()
{
	int v_start, i;
	
	glMatrixMode(GL_TEXTURE);
	glLoadMatrixf(getArcBallRotation());
	glMatrixMode(GL_MODELVIEW);
	
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	
	for (i = 0; i < 6; i++) {
		v_start = i * 4;

		glBegin(GL_QUADS);
		glNormal3f(skybox_verts[v_start+0][0], 
				skybox_verts[v_start+0][1], skybox_verts[v_start+0][2]);
		glVertex3f(skybox_verts[v_start+0][0], 
				skybox_verts[v_start+0][1], skybox_verts[v_start+0][2]);

		glNormal3f(skybox_verts[v_start+1][0], 
				skybox_verts[v_start+1][1], skybox_verts[v_start+1][2]);
		glVertex3f(skybox_verts[v_start+1][0], 
				skybox_verts[v_start+1][1], skybox_verts[v_start+1][2]);

		glNormal3f(skybox_verts[v_start+2][0], 
				skybox_verts[v_start+2][1], skybox_verts[v_start+2][2]);
		glVertex3f(skybox_verts[v_start+2][0], 
				skybox_verts[v_start+2][1], skybox_verts[v_start+2][2]);

		glNormal3f(skybox_verts[v_start+3][0], 
				skybox_verts[v_start+3][1], skybox_verts[v_start+3][2]);
		glVertex3f(skybox_verts[v_start+3][0], 
				skybox_verts[v_start+3][1], skybox_verts[v_start+3][2]);
		glEnd();
	}	
	
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
}

void init_balls()
{
    int i;

    for (i = 0; i < NR_METABALLS; i++)
	{
        balls[i].pos.x = 1.0 + (rand() % (int) (SPACE_SIZE - 2.0));
        balls[i].pos.y = 1.0 + (rand() % (int) (SPACE_SIZE - 2.0));
        balls[i].pos.z = 1.0 + (rand() % (int) (SPACE_SIZE - 2.0));

        balls[i].strength = 0.15;
        balls[i].acc.x = (rand() % 20) * 0.0001;
        balls[i].acc.y = (rand() % 20) * 0.0001;
        balls[i].acc.z = (rand() % 20) * 0.0001;
        balls[i].vel.x = (rand() % 10) * 0.0001;
        balls[i].vel.y = (rand() % 10) * 0.0001;
        balls[i].vel.z = (rand() % 10) * 0.0001;
    }
}


void keyHandler(SDL_keysym *key)
{

    switch (key->sym)
	{
        case SDLK_ESCAPE:
            exit(EXIT_SUCCESS);
            break;
        case SDLK_p:
            if (polygonMode)
			{
                glPolygonMode(GL_FRONT, GL_LINE);
                polygonMode = 0;
            }
            else
			{
                glPolygonMode(GL_FRONT, GL_FILL);
                polygonMode = 1;
            }       
            break; 
        case SDLK_k:
            init_balls();
            break;
            
        default:
            break;
    }
}

void initOpenGL()
{
    glClearDepth(1.0);
    glDepthFunc(GL_LEQUAL);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glEnable(GL_NORMALIZE);
}

void resizeWin(int width, int height)
{
    glViewport(0, 0, (GLsizei) width, (GLsizei) height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45.0, ((GLfloat) width / (height == 0 ? 1 : height)), 
                   0.01, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	setArcBallBounds(width, height);
}

void move_balls()
{
    int i;
    float radius;

    /* Mikachus code :) */

    for (i = 0; i < NR_METABALLS; i++) 
	{
		balls[i].pos.x += balls[i].vel.x;
		if (balls[i].pos.x < 1.5 || balls[i].pos.x > SPACE_SIZE - 1.5)
            balls[i].vel.x = -balls[i].vel.x;
        else
            balls[i].vel.x += balls[i].acc.x;

        balls[i].pos.y += balls[i].vel.y;
        if (balls[i].pos.y < 1.5 || balls[i].pos.y > SPACE_SIZE - 1.5)
            balls[i].vel.y = -balls[i].vel.y;
        else
            balls[i].vel.y += balls[i].acc.y;

        balls[i].pos.z += balls[i].vel.z;
        if (balls[i].pos.z < 1.5 || balls[i].pos.z > SPACE_SIZE - 1.5)
            balls[i].vel.z = -balls[i].vel.z;
        else
            balls[i].vel.z += balls[i].acc.z;
    }
}

void draw(gridcell_t **cells)
{
    int i;
    unsigned int sync;
	
    glClear(GL_DEPTH_BUFFER_BIT);
	glLoadMatrixf(getArcBallRotation());	
	
	render_skybox();
	
	glLoadIdentity();
    glTranslatef(0.0, 0.0, -15.0);
	glMultMatrixf(getArcBallRotation());	
    glTranslatef(-(SPACE_SIZE/2), -(SPACE_SIZE/2), -(SPACE_SIZE/2));
   	
	glUseProgramObjectARB(shaderHandle); 
	//glUniform1fARB(uniformTexUnit, 0);
	
	glBegin(GL_TRIANGLES); 
    for (i = 0; i < NR_METABALLS; i++)    
        make_mesh(cells[i], balls);
    glEnd(); 
    
	glUseProgramObjectARB(0);
    
	if (glGetError())
        printf("%s\n", (const char *) gluErrorString(glGetError()));

    glXGetVideoSyncSGI(&sync);
    glXWaitVideoSyncSGI(2, (sync + 1) % 2, &sync);
   
    SDL_GL_SwapBuffers();
}

int main(int argc, char **argv) {

    SDL_Surface *surface;
    SDL_Event event;
    const SDL_VideoInfo *info;
    int flags;
    char quit = 0;
    gridcell_t **isect;

    srand(time(NULL));
       
    init_balls();
    
    init_grid();

    if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
        fprintf(stderr, "Unable to initialize video: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
   
    atexit(SDL_Quit);
    
    if (!(info = SDL_GetVideoInfo()))
	{
        fprintf(stderr, "Unable to get video info: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
    
    flags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE | SDL_RESIZABLE;

    if (info->hw_available)
	{
        flags |= SDL_HWSURFACE;
        printf("Using hardware surface.\n");
    }
    else
	{
        flags |= SDL_SWSURFACE;
        printf("Using software surface.\n");
    }
    
	if (info->blit_hw)
        flags |= SDL_HWACCEL;

    
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

    if (!(surface = SDL_SetVideoMode(WIDTH, HEIGHT, BPP, flags)))
	{
        fprintf(stderr, "Unable to set video mode: %s\n", SDL_GetError());
        exit(EXIT_FAILURE);
    }
   	
	glewInit();
	
    initOpenGL();
   	load_cube_em();
	setup_shader();
	
    resizeWin(WIDTH, HEIGHT);    
    
    while (!quit) 
	{
        while (SDL_PollEvent(&event))
		{
            switch (event.type)
			{
                case SDL_VIDEORESIZE:
                    if (!(surface = SDL_SetVideoMode(event.resize.w, 
									event.resize.h, BPP, flags)))
					{
                        fprintf(stderr, "Unable to set video mode: %s\n", 
								SDL_GetError());
                        exit(EXIT_FAILURE);
                        return 1;
                    }
                    resizeWin(event.resize.w, event.resize.h);
                    break;
                case SDL_QUIT:
                    quit = 1;
                    break;
                case SDL_KEYDOWN:
                    keyHandler(&event.key.keysym);
                    break;
				case SDL_MOUSEBUTTONDOWN:
					if (event.button.state == SDL_BUTTON(1))
						beginArcBallRotation(event.button.x, event.button.y);
					break;
            	case SDL_MOUSEMOTION:
					if (event.motion.state == SDL_BUTTON(1)) 
						rotateArcBall(event.motion.x, event.motion.y);
					break;
				case SDL_MOUSEBUTTONUP:
						endArcBallRotation();
					break;
			}
        }
       
        isect = find_intersected_cubes(balls);
        draw(isect);
        clean_visited();
        move_balls();
    }
    
    return 0;
}


