#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <GL/glew.h>

#include "shader.h"

char *readCode(const char *file)
{
	static char codebuf[MAX_CODE_LEN];
	ssize_t bytes;
	int fd;
	
	if ((fd = open(file, O_RDONLY)) != -1)
	{
		bytes = read(fd, codebuf, MAX_CODE_LEN - 1);
		codebuf[bytes] = '\0';
	}

	return (fd != -1) ? codebuf : NULL; 
}

int buildErrors(GLuint handle, GLenum errorType)
{
	char log[MAX_LOG_LEN];
	GLint status;
	GLsizei len;

	glGetObjectParameterivARB(handle, errorType,
			&status);
	
	if (!status)
	{
		glGetInfoLogARB(handle, MAX_LOG_LEN - 1, &len, log);
		log[len] = '\0';
		fprintf(stderr, "Error:\n%s\n", log);
	}
	
	return !status;
}

GLuint createShader(GLenum type, const char *file)
{
	GLuint handle = 0;
	char *code;
	
	if ((code = readCode(file)))
	{
		handle = glCreateShaderObjectARB(type);
		glShaderSourceARB(handle, 1, (const char **) &code, NULL);
		glCompileShaderARB(handle);
		
		if (buildErrors(handle, GL_OBJECT_COMPILE_STATUS_ARB))
		{
			glDeleteObjectARB(handle);
			handle = 0;
		}
	}
	
	return handle;
}

GLuint createProgram(GLuint vertexHandle, GLuint fragmentHandle)
{
	GLuint handle;

	handle = glCreateProgramObjectARB();

	glAttachObjectARB(handle, vertexHandle);
	glAttachObjectARB(handle, fragmentHandle);
	glLinkProgramARB(handle);
	
	if (buildErrors(handle, GL_OBJECT_LINK_STATUS_ARB))
	{
		glDeleteObjectARB(handle);
		handle = 0;
	}

	return handle;
}

GLuint createShaderProgram(const char *vertShader, const char *fragShader)
{
	GLuint vHandle, fHandle;

	vHandle = createShader(GL_VERTEX_SHADER_ARB, vertShader);
	fHandle = createShader(GL_FRAGMENT_SHADER_ARB, fragShader);

	return createProgram(vHandle, fHandle); 
}

