/******************************************************************************
 * vectormath.c                                                               *
 *                                                                            *
 * Computer Graphics and Visualization - Fall 05 - Project 3                  *
 *                                                                            *
 * Erik Waling          <c03ewg@cs.umu.se>                                    *
 * December 15 - 2005                                                         *
 ******************************************************************************/

#include <GL/gl.h>
#include <math.h>
#include <string.h>

#include "vectormath.h"

void crossProduct(GLfloat u[3], GLfloat v[3], GLfloat resultVect[3]) 
{
	resultVect[0] = (u[1] * v[2]) - (u[2] * v[1]);
	resultVect[1] = (u[2] * v[0]) - (u[0] * v[2]);
	resultVect[2] = (u[0] * v[1]) - (u[1] * v[0]);
}

GLfloat dotProduct(GLfloat u[3], GLfloat v[3])
{
	return (u[0] * v[0]) + (u[1] * v[1]) + (u[2] * v[2]);
}


GLfloat norm(GLfloat v[3])
{
	return sqrt((v[0] * v[0]) + (v[1] * v[1]) + (v[2] * v[2]));
}

void normalize(GLfloat v[3])
{
	GLfloat n = norm(v);

	v[0] /= n;
	v[1] /= n;
	v[2] /= n;
}

void vectorSub(GLfloat u[3], GLfloat v[3], GLfloat resultVect[3])
{
	resultVect[0] = u[0] - v[0];
	resultVect[1] = u[1] - v[1];
	resultVect[2] = u[2] - v[2];
}

void vectorAdd(GLfloat u[3], GLfloat v[3], GLfloat resultVect[3])
{
	resultVect[0] = u[0] + v[0];
	resultVect[1] = u[1] + v[1];
	resultVect[2] = u[2] + v[2];
}

void vectorAccumAdd(GLfloat u[3], GLfloat v[3])
{
	u[0] = u[0] + v[0];
	u[1] = u[1] + v[1];
	u[2] = u[2] + v[2];
}

void negateVector(GLfloat v[3])
{
	v[0] = -v[0];
	v[1] = -v[1];
	v[2] = -v[2];
}

void vectorScalarDiv(GLfloat v[3], GLfloat scalar)
{
	v[0] /= scalar;
	v[1] /= scalar;
	v[2] /= scalar;
}

void matrixMultiply3(GLfloat a[3][3], GLfloat b[3][3])
{
	GLfloat result[3][3];
	
    result[0][0] = a[0][0] * b[0][0] + a[0][1] * b[1][0] + a[0][2] * b[2][0];
    result[0][1] = a[0][0] * b[0][1] + a[0][1] * b[1][1] + a[0][2] * b[2][1];
    result[0][2] = a[0][0] * b[0][2] + a[0][1] * b[1][2] + a[0][2] * b[2][2];

    result[1][0] = a[1][0] * b[0][0] + a[1][1] * b[1][0] + a[1][2] * b[2][0];
    result[1][1] = a[1][0] * b[0][1] + a[1][1] * b[1][1] + a[1][2] * b[2][1];
    result[1][2] = a[1][0] * b[0][2] + a[1][1] * b[1][2] + a[1][2] * b[2][2];

    result[2][0] = a[2][0] * b[0][0] + a[2][1] * b[1][0] + a[2][2] * b[2][0];
    result[2][1] = a[2][0] * b[0][1] + a[2][1] * b[1][1] + a[2][2] * b[2][1];
    result[2][2] = a[2][0] * b[0][2] + a[2][1] * b[1][2] + a[2][2] * b[2][2];
	
	memcpy(a, result, 9 * sizeof(GLfloat));
}
